"use strict";

var field = new Uint8Array(20400);
var pos = 10200;
var nextPos = 10200;
var dirs = [-101, -1, 101, 1];
var turnDir = Math.random() < 0.5 ? -1 : 1;

var names = ['up', 'left', 'down', 'right'];

function mark(event) {
  if(event.data.done) {
    pos = nextPos;
    field[pos] = 1;
  } else {
    field[nextPos] = 2;
  }
}

function go(event, dir) {
  var offset = dirs[dir];
  nextPos = pos + offset;
  postMessage({id: event.data.id, direction: names[dir]});
}

linePass();

function linePass() {
  var dir = Math.floor(Math.random() * 4);
  var state = 0;
  onmessage = function(event) {
    mark(event);
    if(event.data.done === false || state === 1) {
      dir = (dir + turnDir) & 3;
      if(state === 2) {
        sealPass(dir);
      }
      state++;
    }
    go(event, dir);
  };
}

function sealPass(dir) {
  var targetPos = pos;
  var steps = 0;
  onmessage = function(event) {
    mark(event);
    var done = event.data.done;
    if(done && (++steps % 5)) {
      var p = pos + dirs[(dir - turnDir) & 3];
      if(field[p] === 0) {
        field[p] = 2;
      }
    }
    if(done) {
      if(pos === targetPos) {
        fillPass(dir + turnDir + 1);
      }
      dir = (dir - turnDir) & 3;
    }
    while(field[pos + dirs[dir]] === 2) {
      dir = (dir + turnDir) & 3;
    }
    go(event, dir);
  };
}

function fillPass(dirOffset) {
  var directionBuffer = new Uint16Array(20400);
  var stepCounter = 0;
  var posQueue = new Uint16Array(10000);

  onmessage = function(event) {
    mark(event);
    ++stepCounter;
    posQueue[0] = pos;
    directionBuffer[pos] = stepCounter;
    var i = 0, j = 1;
    var p;
    while(i < j && field[p = posQueue[i++]] !== 0) {
      for(var d = 0; d < 4; ++d) {
        var np = p + dirs[(d + dirOffset) & 3];
        if(field[np] != 2 && (directionBuffer[np] & 0x1fff) !== stepCounter) {
          var dir = (directionBuffer[p] >> 13) || (((d + dirOffset) & 3) + 1);
          directionBuffer[np] = (dir << 13) | stepCounter;
          posQueue[j++] = np;
        }
      }
    }
    go(event, (directionBuffer[p] >> 13) - 1);
  };
}
