"use strict";

var field = new Uint8Array(20400);
var pos = 10200;
var nextPos = 10200;
var dirs = [-101, -1, 101, 1];
var names = ['up', 'left', 'down', 'right'];

function mark(event) {
  if(event.data.done) {
    pos = nextPos;
    field[pos] = 1;
  } else {
    field[nextPos] = 2;
  }
}

function go(event, dir) {
  var offset = dirs[dir];
  nextPos = pos + offset;
  postMessage({id: event.data.id, direction: names[dir]});
}

bouncePass();

function bouncePass() {
  var dir = Math.floor(Math.random() * 4) | 0;
  var x = dir ^ ((dir + 1) & 3);
  var bouncesLeft = 5;
  var stepsLeft = 4;
  onmessage = function(event) {
    mark(event);
    if(!event.data.done) {
      if(--bouncesLeft < 0) {
        sealPass(event, dir);
        return;
      }
      dir ^= 2;
      x ^= 2;
    }
    if(--stepsLeft < 0) {
      stepsLeft = Math.random() * 8;
      dir ^= x;
    }
    go(event, dir);
  };
}

function sealPass(event, dir) {
  var targetPos = pos;
  var turnDir = (Math.floor(Math.random() * 2) * 2 - 1) | 0;
  dir = (dir + turnDir) & 3;
  var stepsLeft = 600;
  go(event, dir);
  onmessage = function(event) {
    mark(event);
    if(event.data.done) {
      if(--stepsLeft < 0 || pos === targetPos) {
        fillPass();
      }
      dir = (dir - turnDir) & 3;
      if(field[pos + dirs[dir]] === 2) {
        dir = (dir + turnDir) & 3;
      }
    }
    else {
      dir = (dir + turnDir) & 3;
    }
    go(event, dir);
  };
}

function fillPass() {
  var distanceBuffer = new Uint32Array(20400);
  var stepCounter = 0;
  var posQueue = new Uint16Array(10000);
  
  onmessage = function(event) {
    mark(event);
    ++stepCounter;
    posQueue[0] = pos;
    distanceBuffer[pos] = stepCounter;
    var i = 0;
    var j = 1;
    var d, p, distance, np;
    while(field[posQueue[i]] !== 0) {
      p = posQueue[i];
      for(d = 0; d < 4; ++d) {
        np = p + dirs[d];
        if(field[np] != 2 && (distanceBuffer[np] & 65535) !== stepCounter){
          distanceBuffer[np] = (j << 16) | stepCounter;
          posQueue[j++] = np;
        }
      }
      ++i;
      if(i >= j) {
        goRandom(event);
        return;
      }
    }
    p = posQueue[i];
    d = 0;
    while(distanceBuffer[p] !== stepCounter) {
      var best = 10000 << 16;
      for(i = 0; i < 4; ++i) {
        distance = distanceBuffer[p + dirs[i]];
        if(distance < best && (distance & 65535) === stepCounter) {
          best = distance;
          d = i ^ 2;
        }
      }
      p += dirs[d ^ 2];
    }
    go(event, d);
  };
}

function goRandom(event) {
  go(event, Math.floor(Math.random() * 4));
  onmessage = goRandom;
}
