"use strict";

var field = new Uint8Array(20400);
var pos = 10200;
var nextPos = 10200;
var lastDir = 0;
var dirs = [-101, -1, 101, 1];
var inc = [0, -1, 0, 1, 0];
var posX = 0, posY = 0, minX = 0, maxX = 0, minY = 0, maxY = 0;

var names = ['up', 'left', 'down', 'right'];

function mark(event) {
  if(event.data.done) {
    pos = nextPos;
    field[pos] = 1;
    if(lastDir & 1) {
      posX += inc[lastDir];
      if(posY === maxY - 99) field[pos - 101] = 2;
      if(posY === minY + 99) field[pos + 101] = 2;
    } else {
      posY += inc[lastDir + 1];
      if(posX === maxX - 99) field[pos - 1] = 2;
      if(posX === minX + 99) field[pos + 1] = 2;
    }
  } else {
    minX = Math.min(minX, posX);
    maxX = Math.max(maxX, posX);
    minY = Math.min(minY, posY);
    maxY = Math.max(maxY, posY);
    field[nextPos] = 2;
  }
}

function go(event, dir) {
  var offset = dirs[dir];
  nextPos = pos + offset;
  lastDir = dir;
  postMessage({id: event.data.id, direction: names[dir]});
}

bouncePass();

function bouncePass() {
  var dir = Math.floor(Math.random() * 4) | 0;
  var x = dir ^ ((dir + 1) & 3);
  var bouncesLeft = 5;
  var stepsLeft = 4;
  onmessage = function(event) {
    mark(event);
    if(field[pos + dirs[dir]] == 2) {
      if(--bouncesLeft < 0) {
        sealPass(event, dir);
        return;
      }
      dir ^= 2;
      x ^= 2;
    }
    if(--stepsLeft < 0) {
      stepsLeft = Math.random() * 8;
      dir ^= x;
    }
    go(event, dir);
  };
}

function sealPass(event, dir) {
  var targetPos = pos;
  var turnDir = (Math.floor(Math.random() * 2) * 2 - 1) | 0;
  dir = (dir + turnDir) & 3;
  go(event, dir);
  var steps = 0;
  onmessage = function(event) {
    mark(event);
    var done = event.data.done;
    if(done && pos === targetPos) {
      fillPass();
    }
    if(done && --steps >= 0) {
      var p = pos + dirs[(dir - turnDir) & 3];
      if(field[p] === 0) {
        field[p] = 2;
      }
    } else {
      steps = 4;
    }
    if(done) {
      if(pos === targetPos) {
        fillPass();
      }
      dir = (dir - turnDir) & 3;
    }
    for(var i = 4; field[pos + dirs[dir]] === 2 && i > 0; --i) {
      dir = (dir + turnDir) & 3;
    }
    go(event, dir);
  };
}

function fillPass() {
  var distanceBuffer = new Uint32Array(20400);
  var stepCounter = 0;
  var posQueue = new Uint16Array(10000);
  var pathStepsLeft = 0;
  
  onmessage = function(event) {
    mark(event);
    ++stepCounter;
    posQueue[0] = pos;
    distanceBuffer[pos] = stepCounter;
    var i = 0;
    var j = 1;
    var d, p, distance, np;
    while(field[posQueue[i]] !== 0) {
      p = posQueue[i];
      for(d = 0; d < 4; ++d) {
        np = p + dirs[d];
        if(field[np] != 2 && (distanceBuffer[np] & 65535) !== stepCounter){
          distanceBuffer[np] = (j << 16) | stepCounter;
          posQueue[j++] = np;
        }
      }
      ++i;
      if(i >= j) {
        goRandom(event);
        return;
      }
    }
    p = posQueue[i];
    d = 0;
    pathStepsLeft = 0;
    while(distanceBuffer[p] !== stepCounter) {
      var best = 10000 << 16;
      var bestDir = 0;
      for(i = 0; i < 4; ++i) {
        distance = distanceBuffer[p + dirs[i]];
        if(distance < best && (distance & 65535) === stepCounter) {
          best = distance;
          bestDir = i;
        }
      }
      p += dirs[bestDir];
      posQueue[pathStepsLeft++] = bestDir ^ 2;
    }
    go(event, posQueue[--pathStepsLeft]);
  };
}

function goRandom(event) {
  go(event, Math.floor(Math.random() * 4));
  onmessage = goRandom;
}
