// http://www.webworkercontest.net
// author: feri
// email: ferenc.hechler@web.de


// ---------------- CODE for reproducible random numbers ----------------- //
var RANDOM_MAX = Math.pow(2,32);
function randomNext() {
	this.seed += (this.seed * this.seed) | 5;
	return (this.seed >>> 32) / RANDOM_MAX;
}
function randomRAND(val) {
	this.seed = ((val!==0)&&(val!==undefined))?val:Math.floor(Math.random()*RANDOM_MAX);
	this.random = randomNext;
}
var WorkerRAND = new randomRAND(1);
//------------------------------------------------------------------------ //

function log(msg) {
	if (typeof console !== 'undefined') {
		console.log(msg);
	}
}

// KONSTANTEN fuer die Richtungen 

var dirMessage	= [ "up", "left", "down", "right", "error"];
var rotateLeft	= [    1,      2,      3,       0,       4];
var rotateRight	= [    3,      0,      1,       2,       4];
var rotateBack 	= [    2,      3,      0,       1,       4];
var addPos 		= [ -101,     -1,    101,       1,       0];
var addX 		= [    0,     -1,      0,       1,       0];
var addY 		= [   -1,      0,      1,       0,       0];


var MODE_RAUTE 			=  1000;
var MODE_RUNNING 		=  2000;
var MODE_CORNER			=  3000;
var MODE_FASTLABYRINTH	=  4000;
var MODE_CIRCLE			=  5000;
var MODE_FILL			=  6000;
var MODE_SEARCHFREE		=  7000;
var MODE_LABYRINTH		=  8000;
var MODE_FASTFILL		=  9000;
var MODE_SPIRAL			= 10000;

// MAIN Variables

var mode = 0;		// aktuelles Programm:	1000 = RAUTE   

var step = 0;		// Aktueller Schritt

var dir = 4;		// zuletzt gesendete Richtung

var posX = 0;		// aktuelle Position
var posY = 0;
var pos = 10201;    // umgerechnet im Array (101+posY)*101+posX

var minX = 0;		// Grenzen der bekannten Welt
var minY = 0;
var maxX = 0;
var maxY = 0;

var borderMinX = 9999;	// Position der bekannten Grenzwaende (9999 wenn unbekannt)
var borderMinY = 9999;
var borderMaxX = 9999;
var borderMaxY = 9999;

var field = new Int16Array(20402);			// Feld, beginn bei Position 10201 (101x101, siehe pos) 
var searchField = new Uint16Array(20402);	// Suchfeld fuer die Wegsuche 


// PRE CALCULATED MOVES

var nextMovesDir = new Int8Array(10000);   	// [-cnt1, dir1, dir2, -cnt3, dir3, ...]  (negative numbers are cnts for the next dir) 
var currentMove = 0;						// currently executed move
var endMove = 0;							// end of pre-claculated move
var lastCalculationPos = 0;					// position, ab der zuletzt der nchste Schritt berechnet wurde

var firstMove = true;

var notifyStepFlag = false;


/**
 * process incoming result.
 */
onmessage = function (event) {

    // Initialisiere Reproduzierbare Zufallszahlen, wenn WorkerRAND.random() statt Math.random() verwendet wird.
//	if (WorkerRAND === undefined) {
//		WorkerRAND = new randomRAND(event.data.randomSeed);
//	}
	
	var unblocked = event.data.done;
	
	if (unblocked) {
		fillStep();
		if (notifyStepFlag) {
			notifyStep();
		}
	}
	else if (firstMove) {
		firstMove = false;
		unblocked = true;
	}
	else {
		fillBlock();
		endMove = 0;
	}

	while (true) {

		if (unblocked) {
			if (currentMove < endMove) {
				doCalculatedMoves();
			}
		}
		else {
			notifyBlocked();
		}
		
		if (!isBlocked(dir)) {
			break;
		}
		unblocked = false;
	}
	
	postMessage({
		id: event.data.id,
		direction: dirMessage[dir]
	});
};



// -------- //
// NOTIFIES //
// -------- //


function notifyStep() {
	switch(mode) {

	// MODE_RUNNING
	case 2000:
		runningStep();
		break;
		
	// MODE_FASTLABYRINTH
	case 4000:
		fastlabyrinthStep();
		break;

	// MODE_CIRCLE
	case 5000:
		circleStep();
		break;
		
	// MODE_FILL
	case 6000:
		fillDoStep();
		break;

	// MODE_LABYRINTH
	case 8000:
		labyrinthStep();
		break;

	// MODE_FASTFILL
	case 9000:
		fastfillDoStep();
		break;

	// MODE_SPIRAL
	case 10000:
		spiralDoStep();
		break;


	}
}

function notifyMoveFinished() {
	switch(mode) {
	
	// MODE_RAUTE
	case 1000:
		rauteMoveFinished();
		break;
		
	// MODE_RUNNING
	case 2000:
		runningMoveFinished();
		break;
		
	// MODE_CORNER
	case 3000:
		cornerMoveFinished();
		break;
		
	// MODE_FASTLABYRINTH
	case 4000:
		fastlabyrinthMoveFinished();
		break;

	// MODE_FILL
	case 6000:
		fillMoveFinished();
		break;

	// MODE_SEARCHFREE
	case 7000:
		searchfreeMoveFinished();
		break;

	// MODE_LABYRINTH
	case 8000:
		labyrinthMoveFinished();
		break;

	// MODE_FASTFILL
	case 9000:
		fastfillMoveFinished();
		break;

	// MODE_SPIRAL
	case 10000:
		spiralMoveFinished();
		break;


	}
}

function notifyBlocked() {
	switch(mode) {

	// MODE_RAUTE
	case 1000:
		rauteBlocked();
		break;
	
	// MODE_RUNNING
	case 2000:
		runningBlocked();
		break;
		
	// MODE_CORNER
	case 3000:
		cornerBlocked();
		break;

	// MODE_FASTLABYRINTH
	case 4000:
		fastlabyrinthBlocked();
		break;
		
	// MODE_CIRCLE
	case 5000:
		circleBlocked();
		break;
		
	// MODE_FILL
	case 6000:
		fillBlocked();
		break;

	// MODE_SEARCHFREE
	case 7000:
		searchfreeBlocked();
		break;

	// MODE_LABYRINTH
	case 8000:
		labyrinthBlocked();
		break;
		
	// MODE_FASTFILL
	case 9000:
		fastfillBlocked();
		break;

	// MODE_SPIRAL
	case 10000:
		spiralBlocked();
		break;


	}
}



// -------------- //
// CALC NEXT MOVE //
// -------------- //

function calcNextMove() {
	// zur Sicherheit alles zuruecksetzen
	clearMoves();
	notifyStepFlag = false;

	switch (mode) {
	// MODE_RAUTE:
	case 1000:
		createFastlabyrinthProgram();
		break;

	// MODE_RUNNING:
	case 2000:
		createFastfillProgram();
		break;

	// MODE_CORNER:
	case 3000:
		createFastfillProgram();
		break;

	// MODE_FASTLABYRINTH:
	case 4000:
		createSearchfreeProgram();
		break;

	// MODE_CIRCLE:
	case 5000:
		createCircleProgram();
		break;

	// MODE_FILL:
	case 6000:
		createSearchfreeProgram();
		break;

	// MODE_SEARCHFREE:
	case 7000:
		createFastfillProgram();
		break;

	// MODE_LABYRINTH:
	case 8000:
		createSearchfreeProgram();
		break;

	// MODE_FASTFILL:
	case 9000:
		createSearchfreeProgram();
		break;

	// MODE_SPIRAL:
	case 10000:
		createFastfillProgram();
		break;

		
	default:
		createCircleProgram();
		break;

	}
//	log("CALCNEXTMOVE "+mode+" ("+posX+", "+posY+") dir="+dir);
}


// --------- //
// FINISHING //
//---------- //

function createFinishingProgram() {
	var cnt = replaceAllUncheckedBlocks();
//	log("replaceAllUncheckedBlocks: "+cnt);
	if (cnt === 0) {
		createCircleProgram();
	}
	else {
		createSearchfreeProgram();
	}
}

function replaceAllUncheckedBlocks() {
	var cnt = 0;
	var pLine = 10201+101*minY+minX;
	var p;
	for (var y=minY; y<=maxY; y++) {
		p = pLine; 
		for (var x=minX; x<=maxX; x++) {
			if (field[p] === -2) {
				// TODO: groesse der zusammenhangenden frei werdenden Flaeche bewerten und sortieren
				field[p] = 0;
				searchField[p] = 0;
				cnt++;
			}
			p++;
		}
		pLine += 101;
	}
	return cnt;
}

// ----------- //
// SPIRAL PROG //
// ----------- //

var spiralRotateFree;
var spiralRotateBlocked;
var spiralStartStep;

function createFillProgram() {
	mode = SPIRAL_FILL;
	notifyStepFlag = true;
	if (Math.random()<0.5) {
		fillRotateFree = rotateLeft;
		fillRotateBlocked = rotateRight;
	}
	else {
		fillRotateFree = rotateRight;
		fillRotateBlocked = rotateLeft;
	}
	fillStartStep = step;
	fillDoStep();
}

function fillDoStep() {
	if (isFree(fillRotateFree[dir])) {
		dir = fillRotateFree[dir];
		return;
	}
	if (isFree(dir)) {
		return;
	}
	if (isFree(fillRotateBlocked[dir])) {
		dir = fillRotateBlocked[dir];
		return;
	}
	if (isFree(rotateBack[dir])) {
		dir = rotateBack[dir];
		return;
	}
	calcNextMove();
}

function fillBlocked() {
	fillDoStep();
}

function fillMoveFinished() {
	fillDoStep();
}



// --------------- //
// SEARCHFREE PROG //
// --------------- //

function createSearchfreeProgram() {
	mode = MODE_SEARCHFREE;
	notifyStepFlag = false;
	
	var bestFirstDi = 4;
	var bestDi = 4;
	var bestCnt = 15;
	var cnt = 0;
	var di = 0;
	var firstDiList = [rotateLeft[di], rotateRight[di], 4];
	
	for (var n=0; n<firstDiList.length; n++) {
		var firstDi = firstDiList[n];
		if (!isBlocked(firstDi)) {
			for (di=0; di<4; di++) {
				cnt = searchFree(di, addPos[firstDi], bestCnt);
				if (firstDi !== 4) {
					cnt ++;
				}
				if (cnt<bestCnt) {
					bestFirstDi = firstDi;
					bestDi = di;
					bestCnt = cnt;
				}
			}
		}
	}
	if (bestDi === 4) {
		fullSearch();
		return;
	}
	if (bestCnt <= 1) {
		calcNextMove();
		return;
	}
	if (bestFirstDi !== 4) {
		createMove(bestFirstDi)
		addRepeatedMove(bestDi, bestCnt-2)
	}
	else {
		createRepeatedMove(bestDi, bestCnt-1);
	}
	startCalculatedMoves();
}


function searchFree(di, poOffs, maxCnt) {
	var addPo = addPos[di];
	var po = pos+poOffs;
	for (var cnt=1; cnt<maxCnt; cnt++) {
		po += addPo;
		if (field[po] === 0) {
			return cnt;
		}
		if (field[po] < 0) {
			return 9999;
		}
	}
	return 9999;
}

var sfStartStep = 1;

function fullSearch() {
//	log("fullSearch ("+posX+"," +posY+")");
	sfStartStep++;
	if (sfStartStep > 64500) {
		searchField = new Uint16Array(20402);
		sfStartStep = 2;
	} 
	var p = 0;
	var currentPosList = [pos];
	searchField[pos] = sfStartStep;
	for (var n=1; n<1000; n++) {
		var nextPosList = [];
		for (var i=0; i<currentPosList.length; i++) {
			for (var di=0; di<4; di++) {
				p = currentPosList[i]+addPos[di];
				if (searchField[p] < sfStartStep) {
					if (field[p] === 0) {
						searchField[p] = sfStartStep+n;
						calcSearchMoves(p, sfStartStep);
						sfStartStep += n+1;
						startCalculatedMoves();
						return;
					}
					if (field[p] < 0) {
						searchField[p] = 65535;
					}
					else {
						searchField[p] = sfStartStep+n;
						nextPosList.push(p);
					}
				}
			}
		}
		if (nextPosList.length === 0) {
			break;
		}
		currentPosList = nextPosList;
	}
	sfStartStep += 1001;
	searchNothingFound = true;
	createFinishingProgram();
}


function calcSearchMoves(startPos, sfStartStep) {
	clearMoves();
	var cnt = searchField[startPos]-sfStartStep;
	var p = startPos;
	var p2 = 0;
	for (var n=cnt-1; n>=0; n--) {
		for (var di=0; di<4; di++) {
			p2=p+addPos[di];
			if (searchField[p2] === sfStartStep+n) {
				nextMovesDir[n] = rotateBack[di];
				p = p2;
				break;
			}
		}
	}
	endMove = cnt-1;
}



function searchfreeBlocked() {
	createCircleProgram();
}

function searchfreeMoveFinished() {
	calcNextMove();
}



// --------- //
// FILL PROG //
// --------- //

var fillRotateFree;
var fillRotateBlocked;
var fillStartStep;

function createFillProgram() {
	mode = MODE_FILL;
	notifyStepFlag = true;
	if (Math.random()<0.5) {
		fillRotateFree = rotateLeft;
		fillRotateBlocked = rotateRight;
	}
	else {
		fillRotateFree = rotateRight;
		fillRotateBlocked = rotateLeft;
	}
	fillStartStep = step;
	fillDoStep();
}

function fillDoStep() {
	if (isFree(fillRotateFree[dir])) {
		dir = fillRotateFree[dir];
		return;
	}
	if (isFree(dir)) {
		return;
	}
	if (isFree(fillRotateBlocked[dir])) {
		dir = fillRotateBlocked[dir];
		return;
	}
	if (isFree(rotateBack[dir])) {
		dir = rotateBack[dir];
		return;
	}
	calcNextMove();
}

function fillBlocked() {
	fillDoStep();
}

function fillMoveFinished() {
	fillDoStep();
}


// ------------- //
// FASTFILL PROG //
// ------------- //

var fafiRotateFree;
var fafiRotateBlocked;
var fafiStartStep;

function createFastfillProgram() {
	mode = MODE_FASTFILL;
	notifyStepFlag = true;
	if (Math.random()<0.5) {
		fafiRotateFree = rotateLeft;
		fafiRotateBlocked = rotateRight;
	}
	else {
		fafiRotateFree = rotateRight;
		fafiRotateBlocked = rotateLeft;
	}
	fafiStartStep = step;
	fastfillDoStep();
}

function fastfillDoStep() {
	if (isFree(fafiRotateFree[dir])) {
		dir = fafiRotateFree[dir];
		return;
	}
	if (isFree(dir)) {
		return;
	}
	if (isFree(fafiRotateBlocked[dir])) {
		dir = fafiRotateBlocked[dir];
		return;
	}
	if (isFree(rotateBack[dir])) {
		dir = rotateBack[dir];
		return;
	}
	calcNextMove();
}

function fastfillBlocked() {
	fastfillDoStep();
}

function fastfillMoveFinished() {
	fastfillDoStep();
}


// ----------- //
// CIRCLE PROG //
// ----------- //

function createCircleProgram() {
	mode = MODE_CIRCLE;
	notifyStepFlag = true;
	circleStep();
}

function circleStep() {
	dir = rotateLeft[dir];
}

function circleBlocked() {
	circleStep();
}


//------------------ //
//FASTLABYRINTH PROG //
//------------------ //

var flabRotateFree;
var flabRotateBlocked;
var flabStartStep;
var flabLastF;
var flabSpeed;
var flabCnt;
var flabTurned;

function createFastlabyrinthProgram() {
	mode = MODE_FASTLABYRINTH;
	notifyStepFlag = true;
	if (Math.random()<0.5) {
		flabRotateFree = rotateLeft;
		flabRotateBlocked = rotateRight;
	}
	else {
		flabRotateFree = rotateRight;
		flabRotateBlocked = rotateLeft;
	}
	flabStartStep = step;
	flabLastF = 0;
	flabSpeed = 2;
	flabCnt = 0;
	flabTurned = false;
	createRepeatedMove(dir, 100);
	startCalculatedMoves();
}

function fastlabyrinthStep() {
	flabTurned = false;
	var currentF = field[pos]; 
	if ((currentF !== step) && (currentF > flabStartStep)) {
		if (flabLastF = currentF-1) {
			calcNextMove();
			return;
		}
	}
	flabLastF = currentF;
	flabCnt++;
	if (flabCnt >= flabSpeed) {
		flabCnt = 0;
		flabSpeed++;
		flabTurned = true;
		dir = flabRotateFree[dir];
	}
	else {
		setUncheckedBlock(flabRotateFree[dir]);
	}
}

function fastlabyrinthBlocked() {
	if (!flabTurned) {
		flabSpeed = 2;
		flabCnt = 0;
	}
	else {
		flabTurned = false;
	}
	dir = flabRotateBlocked[dir];
}

function fastlabyrinthMoveFinished() {
	dir = flabRotateFree[dir];
}


// -------------- //
// LABYRINTH PROG //
// -------------- //

var labyRotateFree;
var labyRotateBlocked;
var labyStartStep;
var labyLastF;

function createLabyrinthProgram() {
	mode = MODE_LABYRINTH;
	notifyStepFlag = true;
	if (Math.random()<0.5) {
		labyRotateFree = rotateLeft;
		labyRotateBlocked = rotateRight;
	}
	else {
		labyRotateFree = rotateRight;
		labyRotateBlocked = rotateLeft;
	}
	labyStartStep = step;
	labyLastF = 0;
	createRepeatedMove(dir, 100);
	startCalculatedMoves();
}

function labyrinthStep() {
	var currentF = field[pos]; 
	if ((currentF !== step) && (currentF > labyStartStep)) {
		if (labyLastF = currentF-1) {
			calcNextMove();
			return;
		}
	}
	labyLastF = currentF;
	dir = labyRotateFree[dir];
}

function labyrinthBlocked() {
	dir = labyRotateBlocked[dir];
}

function labyrinthMoveFinished() {
	dir = labyRotateFree[dir];
}


// ----------- //
// CORNER PROG //
// ----------- //

var cornerVisited = [false, false, false, false];
var cornerPosX = [-100, -100,  100,  100];
var cornerPosY = [-100,  100,  100, -100];

var cornerTargetX = 0;
var cornerTargetY = 0;

function createCornerProgram() {
	mode = MODE_CORNER;
	notifyStepFlag = true;
	
	var c = selectNearestCorner();
	if (c < 0) {
		calcNextMove();
		return;
	}
	cornerTargetX = cornerPosX[c];
	cornerTargetY = cornerPosY[c];
	cornerVisited[c] = true;
	
	createMovesToTarget(cornerTargetX, cornerTargetY);
	startCalculatedMoves();
	return;
}


function cornerMoveFinished() {
	createCornerProgram()
}

function cornerBlocked() {
	createCornerProgram()
}


function selectNearestCorner() {
	var bestDist = 9999;
	var bestC = -1;
	
	for (var c=0; c<4; c++) {
		if (cornerVisited[c]) {
			continue;
		}
		var x = cornerPosX[c];
		var y = cornerPosY[c];
		if (x > minX+99) {
			x = minX+99;
		}
		if (x < maxX-99) {
			x = maxX-99;
		}
		if (y > minY+99) {
			y = minY+99;
		}
		if (y < maxY-99) {
			y = maxY-99;
		}
		var dist = Math.abs(posX-x)+Math.abs(posY-y);
		if (dist == 0) {
			cornerVisited[c] = true;
			continue;
		}
		if (dist < bestDist) {
			bestDist = dist;
			bestC = c;
			cornerPosX[c] = x;
			cornerPosY[c] = y;
		}
	}
	return bestC;
}


// ------------ //
// RUNNING PROG //
// ------------ //


function createRunningProgram() {
	mode = MODE_RUNNING;
	notifyStepFlag = true;
	
	var di = findNextRunningDir(dir);
	createRepeatedMove(di, 100);
	startCalculatedMoves();
}


function runningStep() {
	if (!isFree(dir)) {
		calcNextMove();
	}
}

function runningMoveFinished() {
	calcNextMove()
}

function runningBlocked() {
	calcNextMove()
}


function findNextRunningDir(di) {
	if (isFree(di)) {
		return di;
	}
	var rotate = getRotateLeftOrRight();
	di = rotate[di];
	if (isFree(di)) {
		return di; 
	}
	di = rotate[di];
	if (isFree(di)) {
		return di; 
	}
	di = rotate[di];
	if (isFree(di)) {
		return di; 
	}
	di = rotate[di];
	if (!isBlocked(di)) {
		return di; 
	}
	di = rotate[di];
	if (!isBlocked(di)) {
		return di; 
	}
	di = rotate[di];
	if (!isBlocked(di)) {
		return di; 
	}
	di = rotate[di];
	return di; 
}




// ---------- //
// RAUTE PROG //
// ---------- //

var rauteDir1 = 4;
var rauteDir2 = 4;
var rauteNextWallDirection = 4;
var rauteCntWalls = 0
var rauteCntBounce = 5;

var RAUTE_STEPS = 10;


function createRauteProgram(cntBounce) {
	mode = MODE_RAUTE;
	notifyStepFlag = false;
	rauteCntWalls = 0
	rauteLastWallDirection = 4;
	rauteDir1 = getRandomDirection();	
	rauteDir2 = turnLeftOrRight(rauteDir1);
	rauteCntBounce = cntBounce;
	clearMoves();
	createZigZagMove(rauteDir1, rauteDir2, RAUTE_STEPS, 10);
	startCalculatedMoves();
}

function createZigZagMove(dir1, dir2, len, anz) {
	for (var i=0; i<anz; i++) {
		addRepeatedMove(dir1, len)
		addRepeatedMove(dir2, len)
	}
}

function rauteMoveFinished() {
	calcNextMove();
}

function rauteBlocked() {
	rauteCntWalls++;
	if (rauteCntWalls > 1) {
		if (rauteNextWallDirection !== dir) {
			calcNextMove();
			return;
		}
		if (rauteCntWalls >= rauteCntBounce) {
			calcNextMove();
			return;
		}
	}
	if (dir === rauteDir1) {
		rauteDir1 = rotateBack[rauteDir1];
		createMove(rauteDir1);
		createZigZagMove(rauteDir2, rauteDir1, RAUTE_STEPS, 10);
		rauteNextWallDirection = rauteDir2;
	}
	else {
		rauteDir2 = rotateBack[rauteDir2];
		createMove(rauteDir2);
		createZigZagMove(rauteDir1, rauteDir2, RAUTE_STEPS, 10);
		rauteNextWallDirection = rauteDir1;
	}
	startCalculatedMoves();
}



//-------------------- //
// PRECALCULATED MOVES //
//-------------------- //

function clearMoves() {
	currentMove = 0;
	endMove = 0;
	lastCalculationPos = 0;
}

function createMove(di) {
	currentMove = 0;
	endMove = 1;
	lastCalculationPos = 0;
	nextMovesDir[0] = di;
}

function createRepeatedMove(di, cnt) {
	if (cnt > 0) {
		currentMove = 0;
		endMove = 2;
		lastCalculationPos = 0;
		nextMovesDir[0] = -cnt;
		nextMovesDir[1] = di;
	}
}

function addMove(di) {
	nextMovesDir[endMove] = di;
	endMove++;
}

function addRepeatedMove(di, cnt) {
	if (cnt > 0) {
		nextMovesDir[endMove] = -cnt;
		endMove++;
		nextMovesDir[endMove] = di;
		endMove++;
	}
}


function createMovesToTarget(cornerTargetX, cornerTargetY) {
	var dX = cornerTargetX-posX;
	var dY = cornerTargetY-posY;
	var dirX = 3;
	var dirY = 2;
	if (dX < 0) {
		dX = -dX;
		dirX = rotateBack[dirX];
	}
	if (dY < 0) {
		dY = -dY;
		dirY = rotateBack[dirY];
	}
	if (dX == 0) {
		createRepeatedMove(dirY, dY);
		return;
	}
	if (dY == 0) {
		createRepeatedMove(dirX, dX);
		return;
	}
	
	clearMoves();
	
	var dirL;
	var l;
	var dirS;
	var s;
	if (dX>dY) {
		dirL = dirX;
		l = dX;
		dirS = dirY;
		s = dY;
	}
	else {
		dirL = dirY;
		l = dY;
		dirS = dirX;
		s = dX;
	}
	
	var z = 0;
	for (var i=0; i<s; i++) {
		var n = 0;
		while (z<=l) {
			z += s;
			n++;
		}
		z -= l;
		addRepeatedMove(dirL, n);
		addMove(dirS);
	}
}



function startCalculatedMoves() {
	addMove(4);  // ende marker
	doCalculatedMoves();
}



/**
 * setze "dir" auf die naechste hinterlegte richtung. 
 * gibt es keine weiteren Zuege mehr, 
 * dann wird notifyMoveFinished() aufgerufen.
 */ 
function doCalculatedMoves() {
	if (lastCalculationPos === pos) {
		return;
	}
	lastCalculationPos = pos;
	var di = nextMovesDir[currentMove];
	if (di >= 0) {
//		log("calculated move "+currentMove+": "+di);
		currentMove++;
		if (di === 4) {
			notifyMoveFinished();
		}
		else {
			dir = di;
		}
	}
	else {
		var cnt = di+1;
		dir = nextMovesDir[currentMove+1];
//		log("calculated move "+currentMove+": "+(1-cnt)+"x"+di);
		if (cnt === 0) {
			currentMove = currentMove+2;
		}
		else {
			nextMovesDir[currentMove] = cnt;
		}
	}
}



// --------------- //
// Hilfsfunktionen //
// --------------- //


function getRandomDirection() {
  return Math.floor(Math.random()*4);
}


function turnLeftOrRight(di) {
	return getRotateLeftOrRight()[di];
}

function getRotateLeftOrRight() {
	if (Math.random()<0.5) {
		return rotateLeft;
	}
	return rotateRight;
}


// ----------------- //
// FIELD beschreiben //
// ----------------- //


function initFieldArray() {
	posX = 0;
	posY = 0;
	pos = 10201;
	step = 1;
	field[pos] = step;
}

function fillStep() {
	step++;
	switch (dir) {
	case 0:
		posY -= 1;
		pos -= 101;
		if (field[pos]===0) {
			field[pos] = step;
		}
		break;
	case 1:
		posX -= 1;
		pos -= 1;
		if (field[pos]===0) {
			field[pos] = step;
		}
		break;
	case 2:
		posY += 1;
		pos += 101;
		if (field[pos]===0) {
			field[pos] = step;
		}
		break;
	case 3:
		posX += 1;
		pos += 1;
		if (field[pos]===0) {
			field[pos] = step;
		}
		break;
	default: 
		break;
	}
}

function fillBlock() {
	switch (dir) {
	case 0:
		field[pos-101] = -1;
		if (posY < minY) {
			minY = posY;
			if (maxY-minY === 99) {
				borderMinY = minY;
				borderMaxY = maxY;
			}
		}
		break;
	case 1:
		field[pos-1] = -1;
		if (posX < minX) {
			minX = posX;
			if (maxX-minX === 99) {
				borderMinX = minX;
				borderMaxX = maxX;
			}
		}
		break;
	case 2:
		field[pos+101] = -1;
		if (posY > maxY) {
			maxY = posY;
			if (maxY-minY === 99) {
				borderMinY = minY;
				borderMaxY = maxY;
			}
		}
		break;
	case 3:
		field[pos+1] = -1;
		if (posX > maxX) {
			maxX = posX;
			if (maxX-minX === 99) {
				borderMinX = minX;
				borderMaxX = maxX;
			}
		}
		break;
	default: 
		break;
	}
}

function setUncheckedBlock(di) {
	var p = pos+addPos[di];
	if (field[p] === 0) {
		field[p] = -2;
	}
}


// ----------- //
// FIELD lesen //
// ----------- //


function isFree(di) {
	return (field[pos+addPos[di]] === 0);
}

function isOwn(di) {
	return (field[pos+addPos[di]] > 0);
}

function isBlocked2(di1, di2) {
	return (field[pos+addPos[di1]+addPos[di2]] < 0);
}

function isBlocked(di) {
	var f = field[pos+addPos[di]];
	if (f === 0) {
		switch(di) {
		case 0:
			if (borderMinY === posY) {
				field[pos+addPos[di]] = -1;
				f = -1;
			}
			break;
		case 1:
			if (borderMinX === posX) {
				field[pos+addPos[di]] = -1;
				f = -1;
			}
			break;
		case 2:
			if (borderMaxY === posY) {
				field[pos+addPos[di]] = -1;
				f = -1;
			}
			break;
		case 3:
			if (borderMaxX === posX) {
				field[pos+addPos[di]] = -1;
				f = -1;
			}
			break;
		}
	}
	return (f < 0);
}


// ---- //
// INIT //
//---- //


init();

function init() {
	initFieldArray();
	createRauteProgram(5);
}



